affinityMatrix<-function(Diff,sigma=1){
  densities = dnorm(Diff,0,sigma*sigma*2,log = FALSE)
  
  W = (densities + t(densities)) / 2
  diag(W) = 1
  return(W)
}

bigraph_partition<-function(goodluck,alt=1){
  W = getW(goodluck)  
  realK = estimateNumberOfClustersGivenGraph(affinityMatrix(W))[[alt]]
  q1 = estimateNumberOfClustersGivenGraph(affinityMatrix(W))[[alt+2]]
  #sc = specc(W,centers=realK)
  #group = (sc@.Data[(ncol(goodluck)+1):nrow(W)])
  wsm <- spectralClustering(W,realK)[(ncol(goodluck)+1):nrow(W)]
  return(list(wsm=wsm,q1=q1))
}
bigraph_partition_s<-function(goodluck,alt=1){
  W = affinityMatrix(as.matrix(dist(goodluck)))
  realK = estimateNumberOfClustersGivenGraph(affinityMatrix(W))[[alt]]
  q1 = estimateNumberOfClustersGivenGraph(affinityMatrix(W))[[alt+2]]
  #sc = specc(W,centers=realK)
  #group = (sc@.Data[(ncol(goodluck)+1):nrow(W)])
  wsm <- spectralClustering(W,realK)[(ncol(goodluck)+1):nrow(W)]
  return(list(wsm=wsm,q1=q1))
}
crEnsemble<-function(X,M,k,scheme){
  E = NULL
  klist = NULL
  if (scheme==1){
    for (i in 1:M){
      Cl = kmeans(X,k,iter.max = 100)$cluster
      while(length(unique(Cl))!=k){
        Cl = kmeans(X,k,iter.max = 100)$cluster
      }
      E = cbind(E,Cl)
    }
  }else{
    for (i in 1:M){
      K = floor((k-2+1)*runif(1) + 2)
      klist = c(klist,K)
      Cl = kmeans(X,K,iter.max = 100)$cluster
      while(length(unique(Cl))!=K){
        Cl = kmeans(X,K,iter.max = 100)$cluster
      }
      E = cbind(E,Cl)
    }
  }
  ri = NULL
  #for(i in 1:ncol(E)){
   # ri = c(ri,randIndex(gnd,E[,i]))
  #}
  return(list(E=E,klist=klist))#,ri=ri))
}
crEnsemble_bagging<-function(X,M,k,scheme){
  E = NULL
  klist = NULL
  if (scheme==1){
    for (i in 1:M){
      Cl = kmeans_bagging(X,k)
      while(length(unique(Cl))!=k){
        Cl = kmeans_bagging(X,k)
      }
      E = cbind(E,Cl)
    }
  }else{
    for (i in 1:M){
      K = floor((k-2+1)*runif(1) + 2)
      klist = c(klist,K)
      Cl = kmeans_bagging(X,K)
      while(length(unique(Cl))!=K){
        Cl = kmeans_bagging(X,K)
      }
      E = cbind(E,Cl)
    }
  }
  ri = NULL
  for(i in 1:ncol(E)){
    ri = c(ri,randIndex(gnd,E[,i]))
  }
  return(list(E=E,klist=klist,ri=ri))
}
crEnsemble_nolabel<-function(X,M,k,scheme){
  E = NULL
  klist = NULL
  if (scheme==1){
    for (i in 1:M){
      Cl = kmeans(X,k,iter.max = 100)$cluster
      while(length(unique(Cl))!=k){
        Cl = kmeans(X,k,iter.max = 100)$cluster
      }
      E = cbind(E,Cl)
    }
  }else{
    for (i in 1:M){
      K = floor((k-2+1)*runif(1) + 2)
      klist = c(klist,K)
      Cl = kmeans(X,K,iter.max = 100)$cluster
      while(length(unique(Cl))!=K){
        Cl = kmeans(X,K,iter.max = 100)$cluster
      }
      E = cbind(E,Cl)
    }
  }
  return(list(E=E,klist=klist))
}
estimateNumberOfClustersGivenGraph<-function(W, NUMC=2:10) {
  if (min(NUMC) == 1) {
    warning('Note that we always assume there are more than one cluster.');
    NUMC = NUMC[NUMC > 1]  
  }
  
  W = (W + t(W))/2
  diag(W) = 0
  
  if (length(NUMC) > 0) {
    degs = rowSums(W)
    
    
    # compute unnormalized Laplacian
    
    degs[degs == 0] = .Machine$double.eps    
    D = diag(degs)    
    L = D - W
    Di = diag(1 / sqrt(degs))
    L = Di %*% L %*% Di
    
    # compute the eigenvectors corresponding to the k smallest
    # eigs$valuess
    eigs = eigen(L)
    eigs_order = sort(eigs$values, index.return=T)$ix
    eigs$values = eigs$values[eigs_order]
    eigs$vectors = eigs$vectors[, eigs_order]
    eigengap = abs(diff(eigs$values))
    eigengap = eigengap * (1 - eigs$values[1:length(eigs$values) - 1] ) / (1 - eigs$values[2:length(eigs$values)])
    
    quality = list()
    for (c_index in 1:length(NUMC)) {
      ck = NUMC[c_index]
      UU = eigs$vectors[, 1:ck]
      EigenvectorsDiscrete <- .discretisation(UU)[[1]]
      EigenVectors = EigenvectorsDiscrete^2
      
      # MATLAB: sort(EigenVectors,2, 'descend');
      temp1 <- EigenVectors[do.call(order, lapply(1:ncol(EigenVectors), function(i) EigenVectors[, i])), ]
      temp1 <- t(apply(temp1, 1, sort, TRUE))  
      
      quality[[c_index]] = (1 - eigs$values[ck + 1]) / (1 - eigs$values[ck]) * 
        sum( sum( diag(1 / (temp1[, 1] + .Machine$double.eps) ) %*% temp1[, 1:max(2, ck-1)] ))
    }
    
    t1 <- sort(eigengap[NUMC], decreasing=TRUE, index.return=T)$ix
    K1 = NUMC[t1[1]]
    K12 = NUMC[t1[2]]
    t2 <- sort(unlist(quality), index.return=TRUE)$ix
    K2 <- NUMC[t2[1]]
    K22 <- NUMC[t2[2]]    
  }
  q1 = sort(eigengap[NUMC], decreasing=TRUE, index.return=T)$x[1]
  q2 = sort(unlist(quality), index.return=TRUE)$x[1]
  return (list(K1, K12,q1, K2, K22,q2))
}
extend<-function(E){
  #E = relabelCl(E)$newE
  N = nrow(E)
  no_allcl = max(E)
  pc = matrix(0,nrow = N,ncol=no_allcl)
  for (i in 1:N){
    pc[i,E[i,]] = 1
  }
  return(pc)
}
findMaxHeight<-function (hc, maxK) {
  height <- rev(diff(hc$height))[1:(maxK-1)]
  i=which.max(height)
  max(height)
  i+1
  return(list((i+1),max(height)))
}
getAverageCmatrix<-function(CM,n){
  ite <- length(CM)
  s <- matrix(0,nrow=n,ncol=n)
  for(i in 1:ite){
    s <- s+CM[[i]]
  }
  s <- s/ite
  return(s)
}
getCoMatrix<-function(E){
  CM = list()
  for (i in 1:ncol(E)){
    CM[[i]] = getConnectivityMatrix(E[,i],nrow(E))
  }
  AverageCmatrix = getAverageCmatrix(CM,nrow(E))
  return(AverageCmatrix)
}
getConnectivityMatrix<-function(group,N)
{
  clus <- length(unique(group))
  S=matrix(0,N,N)
  
  for (j in 1:clus){
    X=rep(0,N);
    X[which(group==j)]=1
    S=S+X%*%t(X)
  }
  return(S)
}
getW<-function(A){
  n = nrow(A)
  t = ncol(A)
  #print(n)
  #print(t)
  W <- matrix(0,nrow=(n+t),ncol=(n+t))
  W[(t+1):(t+n),1:t] <- A
  W[(1:t),(t+1):(t+n)] <- t(A)
  return(W)
}
hclust_cc<-function(goodluck,hNum=1){
  hc = switch(hNum,hclust(dist(goodluck),method = "complete"),hclust(dist(goodluck),method = "average"),
              hclust(dist(goodluck),method = "single"))
  realK = findMaxHeight(hc,10)[1]
  wsm = cutree(hc,realK)
  q1 = findMaxHeight(hc,10)[2]
  return(list(wsm=wsm,q1=q1))
}
iter_last_duo<-function(E,ithreshold=1e-5,threshold=1e-5,alfa=0.7){
  obj_net = getCoMatrix(E)
  E_extend = extend(E_relabel)
  rate = 1
  clu_net_rate = 1
  obj_net_rate = 1
  on_obj_net = smooth(obj_net,t(E_extend),threshold,alfa)
  clu_net = affinityMatrix(as.matrix(dist(on_obj_net)))
  while((clu_net_rate > ithreshold)|(obj_net_rate > ithreshold))
  {
    on_clu_net = smooth(clu_net,E_extend,threshold,alfa)
    obj_net_new = affinityMatrix(as.matrix(dist(on_clu_net)))
    obj_net_rate = norm((obj_net_new-obj_net),"1")
    obj_net = obj_net_new
    on_obj_net = smooth(obj_net_new,t(E_extend),threshold,alfa)
    clu_net_new = affinityMatrix(as.matrix(dist(on_obj_net)))
    clu_net_rate = norm((clu_net_new-clu_net),"1")
    clu_net = clu_net_new
    #print(obj_net_rate)
    #print(clu_net_rate)
  }
  #on_clu_net = smooth(clu_net,t(on_obj_net))
  return(list(clusim = clu_net,goodluck=on_clu_net,objsim=obj_net))
}
jaccard<-function(a,b){
  M11 = sum(a * b)
  M10M01 = sum((a+b)==1)
  J = M11/(M10M01+M11)
  return(J)
}
Jmatrix<-function(pc){
  jmatrix = matrix(0,nrow=ncol(pc),ncol=ncol(pc))
  for(i in 1:ncol(pc)){
    for(j in 1:ncol(pc)){
      jmatrix[i,j] = jaccard(pc[,i],pc[,j])
    }
  }
  #goodluck = smooth(jmatrix,E_extend,threshold,alpha)
  return(jmatrix)
  
}
my_survival<-function(groupsNew){
  coxFit <- coxph(Surv(time = Survival, event = Death) ~ as.factor(groupsNew), data = survival, ties="exact")
  mfit <- survfit(Surv(Survival, Death == 1) ~ as.factor(groupsNew), data = survival)
  return(round(summary(coxFit)$sctest[3],digits = 5))
}
normalize1<-function(X){
  for(i in 1:length(X[1,]))
  {
    chu <- sum(X[,i])
    X[,i] <- X[,i]/chu
  }
  return(X)
}
relabelCl<-function(E){
  N = nrow(E);
  M = ncol(E);
  newE = matrix(0,nrow=N,ncol=M);
  #--- first clustering
  ucl = unique(E[,1]);
  if(max(E[,1])!=length(ucl)){
    for (j in 1:length(ucl)){
      newE[(E[,1]==ucl[j]),1] = j;
    }
  }else{
    newE[,1] = E[,1];
  }
  #--- the rest of the clustering
  for(i in 2:M){
    ucl = unique(E[,i]);
    prevCl = 0;
    #prevCl = max(newE[,1:i-1])
    for(k in 1:(i-1)){
      prevCl = prevCl + length(unique(E[,k]))
    }
    for (j in 1:length(ucl)){
      newE[(E[,i]==ucl[j]),i] = prevCl + j;
    }
  }
  no_allcl = max(newE)
  return(list(no_allcl=no_allcl,newE=newE))
}
RWCE_main<-function(E,alpha=0.3,gNum=1,scaleNum=1,ccNum=1,bipNum=1,hNum=1,iter=5){
  result = list()
  quality = list()
  E_relabel = relabelCl(E)$newE
  E_extend = extend(E_relabel)
  jmatrix = Jmatrix(E_extend)
  gmatrix = smooth(jmatrix,E_extend,1e-3,alpha)
  g_scale_matrix = scaleusingclusim(E,E_extend,gmatrix,1)
  Result = bigraph_partition(g_scale_matrix,bipNum)
  result = Result$wsm
  quality = Result$q1
  group = result
  index = quality
  kei = length(unique(group))
  return(list(group=group,index=index,k=kei))
}
scaleusingclusim<-function(E,E_extend,gmatrix,dc=0.8){
  clustersim = affinityMatrix(as.matrix(dist(t(gmatrix))))
  E_extend_new = E_extend
  aE = apply(E,2,max)
  for (i in 1:length(aE)){
    st = (sum(aE[1:i-1])+1)
    en = ((sum(aE[1:i-1])+1)+aE[i]-1)
    for(j in 1:nrow(E_extend)){
      clusterbelong = which(E_extend[j,st:en]==1)+st-1
      fenmu = sum(clustersim[clusterbelong,which(E_extend[j,st:en]!=1)+st-1])+(.Machine$double.eps)
      for(z in (which(E_extend[j,st:en]!=1)+st-1)){
        E_extend_new[j,z] = dc*(clustersim[clusterbelong,z]/fenmu)
      }
    }
  }
  return(E_extend_new)
}
scaling_direct<-function(E,E_extend,goodluck,dc = 0.8){
  E_extend_new = E_extend
  aE = apply(E,2,max)
  for (i in 1:length(aE)){
    st = (sum(aE[1:i-1])+1)
    en = ((sum(aE[1:i-1])+1)+aE[i]-1)
    for(j in 1:nrow(E_extend)){
      clusterbelong = which(E_extend[j,st:en]==1)+st-1
      fenmu = sum(goodluck[j,which(E_extend[j,st:en]!=1)+st-1])+(.Machine$double.eps)
      for(z in (which(E_extend[j,st:en]!=1)+st-1)){
        E_extend_new[j,z] = dc * (goodluck[j,z] / fenmu)
      }
    }
    #jnew[st:en,st:en][jnew[st:en,st:en]==0] = meanJ
  }
  return(E_extend_new)
}
smooth<-function(network,F0,threshold=1e-5,alfa=0.7){
  diag(network) = 1
  A <- normalize1(network)
  rate <- 1
  F.old <- F0
  iter = 0
  while((rate>threshold)&&(iter<100)){
    iter = iter + 1
    F.new <- alfa*F.old%*%A +(1-alfa)*F0
    rate <- norm((F.new-F.old),"1")
    F.old <- F.new
  }
  return (F.new)
}
smooth_withupdatednetwork<-function(network,F0,threshold=1e-5,alfa=0.7){
  diag(network) = 1
  A <- normalize1(network)
  rate = 1
  A.rate = 1
  F.old = F0
  A.iter = 0
  iter = 0
  while((A.rate > threshold)&&(A.iter<100))
  {
    A.iter = A.iter + 1
    while((rate>threshold)&&(iter<100)){
      iter = iter + 1
      F.new <- alfa*F.old%*%A +(1-alfa)*F0
      rate <- norm((F.new-F.old),"1")
      F.old <- F.new
    }
    A_new <- affinityMatrix(as.matrix(dist(t(F.old))))
    A_new <- normalize1(A_new)
    A.rate <- norm((A_new-A),"1")
    A = A_new
    rate = 1
    F.old = F0
  }
  #print(A.iter)
  #print(iter)
  return(F.new)
}
